/*
 * Copyright 2019 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "base/status.h"

#include "absl/strings/str_cat.h"
#include "absl/types/optional.h"

namespace cel_base {

std::string StatusCodeToString(StatusCode code) {
  switch (code) {
    case StatusCode::kOk:
      return "OK";
    case StatusCode::kCancelled:
      return "CANCELLED";
    case StatusCode::kUnknown:
      return "UNKNOWN";
    case StatusCode::kInvalidArgument:
      return "INVALID_ARGUMENT";
    case StatusCode::kDeadlineExceeded:
      return "DEADLINE_EXCEEDED";
    case StatusCode::kNotFound:
      return "NOT_FOUND";
    case StatusCode::kAlreadyExists:
      return "ALREADY_EXISTS";
    case StatusCode::kPermissionDenied:
      return "PERMISSION_DENIED";
    case StatusCode::kUnauthenticated:
      return "UNAUTHENTICATED";
    case StatusCode::kResourceExhausted:
      return "RESOURCE_EXHAUSTED";
    case StatusCode::kFailedPrecondition:
      return "FAILED_PRECONDITION";
    case StatusCode::kAborted:
      return "ABORTED";
    case StatusCode::kOutOfRange:
      return "OUT_OF_RANGE";
    case StatusCode::kUnimplemented:
      return "UNIMPLEMENTED";
    case StatusCode::kInternal:
      return "INTERNAL";
    case StatusCode::kUnavailable:
      return "UNAVAILABLE";
    case StatusCode::kDataLoss:
      return "DATA_LOSS";
    default:
      return "";
  }
}

std::ostream& operator<<(std::ostream& os, StatusCode code) {
  return os << StatusCodeToString(code);
}

Status::Status(StatusCode code, absl::string_view message)
    : code_(code), message_(code == StatusCode::kOk ? "" : message) {}

std::string Status::ToString() const {
  return ok() ? "OK" : absl::StrCat(StatusCodeToString(code()), ": ", message_);
}

void Status::SetPayload(absl::string_view type_url, const StatusCord& payload) {
  if (!ok()) {
    payload_.try_emplace(std::string(type_url), payload);
  }
}

absl::optional<StatusCord> Status::GetPayload(
    absl::string_view type_url) const {
  auto it = payload_.find(std::string(type_url));
  if (it == payload_.end()) return absl::nullopt;
  return it->second;
}

void Status::ErasePayload(absl::string_view type_url) {
  payload_.erase(std::string(type_url));
}

std::ostream& operator<<(std::ostream& os, const Status& x) {
  return os << x.ToString();
}

}  // namespace cel_base
